import { RollStats } from "./rollstats"
import { Outcomes } from "./asphodice"
import palette from "google-palette"
import { Chart } from "chart.js"

let barChartOptions = { scales: {
				yAxes: [{
					ticks: {
						beginAtZero: true,
					}
				}]
			},
			legend: {
				display: false,
			}
}

let rollstats = new RollStats();

function hexColours(numColours: number): Array<string> {
	return palette("cb-Set1", numColours).map( function (hex: string) {
		return `#${String(hex)}`; })
}
/**
 * Make a table with headings
 *
 * TODO: boolean for striped
 */
function makeTableSkeleton(tableId: string, headings: Array<string>) {
	let output = `<table id="${tableId}" class="table table-striped">
	<thead>`;
	for (let head of headings) {
		output += `<th>${head}</th>`;
	}
	output += "</thead><tbody></tbody></table>";
	return output;
}

/**
 * Remap true to "rerolled" and false to "Not rerolled"
 */
function mapRerolledKeys(Keys: Array<string>): Array<string> {
	return Keys.map((k) => {return (k == "true") ? "Rerolled" : "Not Rerolled"});
}

function buildData(): string {
	let output = "";
	return output;
}

function buildCharts(): string {
	let output = "";
	return output;
}

/**
 * h2 wrapper
 */
function bigTitle(text:string): string {
	return `<h2 class="card-title">${text}</h2>`;
}

/**
 * h3 wrapper
 */
function subTitle(text:string): string {
	return `<h3 class="card-subtitle my-1">${text}</h3>`;
}

/**
 * h4 wrapper
 */
function subsubTitle(text:string): string {
	return `<h4 class="card-subtitle my-1">${text}</h4>`;
}

/**
 * Report on rerolls
 * ie:
 *  - make chart
 *  - make number boxes
 */

function rerollReport(rollstats: RollStats): string {
	let output = subTitle("Rerolls");
	// Preamble (ie it looks super confusing)
	output += "<p>This is super confusing because the colours are swapped on the legend and the chart, then the data below swaps back.</p>";
	// Chart
	output += `<canvas id="rerollsChart"></canvas>`

	// Numbers

	output += `<div class="row my-3 mx-auto">`;
	output += `<div class="col alert alert-success">
			<strong>Rerolled:</strong>
			${rollstats.rerollCounts.true}
			(${(rollstats.rerollCounts.true / rollstats.numRolls * 100).toFixed(2)} %)
			</div>`;
	output += `<div class="col alert alert-warning">
			<strong>Not rerolled:</strong>
			${rollstats.rerollCounts.false}
			(${(rollstats.rerollCounts.false / rollstats.numRolls * 100).toFixed(2)} %)
			</div>`
	output += `</div>`
	return output;
}

function outcomesReport(rollstats: RollStats): string {
	let output = subTitle("Outcomes");
	// Preamble

	// Chart
	output += `<canvas id="outcomesChart"></canvas>`;

	// Table
	output += makeTableSkeleton("outcomeTable", ["Outcome", "Count", "Percentage"]);
	return output;
}

function outcomeBalancesReport(rollstats: RollStats): string {
	let output = subTitle("Outcome Balances");
	// Preamble

	// Chart
	output += `<canvas id="balanceChart"></canvas>`;

	// Table
	output += makeTableSkeleton("balanceTable", ["Balance", "Count", "Percentage"]);
	return output;
}

/**
 * Do the charts after we've added them to the canvas
 *
 * The more javascripty way of doing this would be a callback or a custom event (TODO?)
 */
function generateCharts(rollstats: RollStats): void {

	/*
	 * Rerolls
	 */

	let rerollsCanvas: any = $("#rerollsChart");
	let rerollsChart = new Chart(rerollsCanvas, {
		type: "doughnut",
		data: {
			labels: mapRerolledKeys(Object.keys(rollstats.rerollCounts)),
			datasets: [{
				label: "Reroll Counts",
				data: Object.values(rollstats.rerollCounts),
				backgroundColor: ["#d1e7dd", "#fff3cd"],
				//backgroundColor: hexColours(Object.values(rollstats.balanceCounts).length),
			}],
		},
		options: {
			title: {
				text: "Rerolls Chart",
				display: false,
			},
			legend: {
				reverse: true,
			}
		}
	});

	/*
	 * Outcomes
	 */

	let outcomesCanvas: any = $("#outcomesChart");
	let outcomesChart = new Chart(outcomesCanvas, {
		type: "bar",
		data: {
			labels: Object.keys(rollstats.outcomeCounts),
			datasets: [{
				label: "Outcome Counts",
				data: Object.values(rollstats.outcomeCounts),
				backgroundColor: hexColours(Object.values(rollstats.balanceCounts).length),
			}],
		},
		options: barChartOptions,
	});

	/**
	 * Outcome Balances
	 */

	let bc = rollstats.balanceCounts;
	let keys = Object.keys(bc);
	keys.sort(function(a: string, b: string){return Number(a) - Number(b)});
	// sort values too
	let values = [];
	for (let i = 0; i < keys.length; i++) {
		values.push(rollstats.balanceCounts[keys[i]]);
	}

	let balanceCanvas: any = $("#balanceChart");
	let balanceChart = new Chart(balanceCanvas, {
		type: "bar",
		data: {
			labels: keys,
			datasets: [{
				label: "Outcome Balance Counts",
				data: values,
				backgroundColor: hexColours(Object.values(rollstats.balanceCounts).length),
			}],
		},
		options: barChartOptions,
	});

}

/**
 * Generate tables after we've inserted them into the DOM,
 *
 * Also should be callback/event-driven. (TODO?)
 */
function generateTables(rollstats: RollStats): void {
	/*
	 * Outcomes
	 */

	let oc = rollstats.outcomeCounts;
	let okeys = Object.keys(oc);
	for (var i = 0; i < okeys.length; i++) {
		let tb = $("#outcomeTable").find("tbody");
		let outcome: Outcomes = okeys[i] as Outcomes;
		let outcomeCount = rollstats.outcomeCounts[outcome];
		let outcomePercent: string = "";
		if (outcomeCount) {
			// Needed to avoid TS2532:
			outcomePercent = (outcomeCount / rollstats.numRolls * 100).toFixed(2);
		}
		tb.append(`<tr> <td>${outcome}</td>
			  <td>${outcomeCount}</td>
			  <td>${outcomePercent}</td>
			  </tr>`);
	}

	/*
	 * Outcome Balance Counts
	 */

	let bc = rollstats.balanceCounts;
	let keys = Object.keys(bc);
	keys.sort(function(a: string, b: string){return Number(a) - Number(b)});
	for (var i = 0; i < keys.length; i++) {
		let tb = $("#balanceTable").find("tbody");
		tb.append(`<tr>	<td>${keys[i]}</td>
				<td>${rollstats.balanceCounts[keys[i]]}</td>
				<td>${(rollstats.balanceCounts[keys[i]] / rollstats.numRolls * 100).toFixed(2)}</td>
			</tr>`);
	}

}

/**
 * For pretty-printing. See https://stackoverflow.com/a/2901298
 */
function numberWithCommas(x:number): string {
	return x.toString().replace(/\B(?=(\d{3})+(?!\d))/g, ",");
}

/**
 * Give some info on the dice used to generate these results
 */

function describeRolls(rollstats: RollStats): string {
	let output = subTitle("Description");
	output += `<p>There were <strong>${rollstats.numDice}</strong> Asphodice rolled <strong>${numberWithCommas(rollstats.numRolls)}</strong> times.</p>`;
	return output;
}

function getResults():void {
	// Disable 'roll' button (TODO: multiple sets of rolls)
	$("#mainRoll").prop("disabled", true);
	console.log("Getting results..."); // TODO: progress bar when we have multiple sets of rolls

	$("#results").empty();
	rollstats.doRolls();

	$("#results").append(`<div class="card my-4">`);
	let resultsCard = $("#results").find("div.card");
	resultsCard.append(`<div class="card-body">`);
	let resultsCardBody = $("#results").find("div.card-body");


	resultsCardBody.append(bigTitle("Asphodice Results"));
	resultsCardBody.append(describeRolls(rollstats));
	resultsCardBody.append(rerollReport(rollstats));
	resultsCardBody.append(outcomesReport(rollstats));
	resultsCardBody.append(outcomeBalancesReport(rollstats));

	// Post-DOM-Construction generation
	//
	// For tables: we need this as the way that tables were implemented
	// we .find() the <tbody>
	//
	// For charts: the Chart() generation function expects to passed an extant
	// <canvas> element

	// Tables
	generateTables(rollstats);

	// Charts
	generateCharts(rollstats);
}


function setupHandlers(): void {
	$("#mainRoll").click(getResults);
}

document.addEventListener("DOMContentLoaded", setupHandlers);
